package it.davideorlandi.jtbtar;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import org.apache.commons.io.IOUtils;

import com.ice.tar.TarEntry;
import com.ice.tar.TarOutputStream;

/**
 * Easy object for creating tar files.
 * @author Davide
 */
public class EasyTar
{
    private static final int BUFFER_SIZE = 1024 * 1024;

    private File baseDirectory;

    private File destinationFile;

    private OutputStream fos;

    private BufferedOutputStream bos;

    private TarOutputStream tos;

    /**
     * Constructor.
     * @param baseDirectory base source directory. Source file names will be relative to this
     *        location.
     * @param destinationFile destination tar archive.
     * @throws IOException error writing the archive.
     */
    public EasyTar(final File baseDirectory, final File destinationFile) throws IOException
    {
        this.baseDirectory = baseDirectory;
        this.destinationFile = destinationFile;

        boolean ok = false;
        try
        {
            fos = new FileOutputStream(this.destinationFile);
            bos = new BufferedOutputStream(fos, BUFFER_SIZE);
            tos = new TarOutputStream(bos);
            ok = true;
        }
        finally
        {
            // close streams if partially open
            if ( ! ok)
                close();
        }

    }

    /**
     * Adds a file to the archive.
     * @param relativeFileName file name relative to the base directory.
     * @throws IOException read or write error.
     */
    public void addFile(final String relativeFileName) throws IOException
    {
        File absoluteFile = new File(baseDirectory, relativeFileName);
        if ( ! absoluteFile.exists())
        {
            throw new IOException("File not found: " + absoluteFile);
        }
        TarEntry entry = new TarEntry(absoluteFile);
        if (absoluteFile.isFile())
        {
            entry.setSize(absoluteFile.length());
            entry.setName(relativeFileName);
            tos.putNextEntry(entry);
            FileInputStream fis = new FileInputStream(absoluteFile);
            try
            {
                IOUtils.copy(fis, tos);
            }
            finally
            {
                IOUtils.closeQuietly(fis);
            }
        }
        else
        {
            entry.setSize(0);
            entry.setName(relativeFileName + "/");
            tos.putNextEntry(entry);
        }

        tos.closeEntry();
    }

    /**
     * Adds an entry mad from a string to the archive.
     * @param relativeFileName virtual file name.
     * @param content content to be written.
     * @param charset encoding for the string, typically US-ASCII or ISO-8859-1.
     * @throws IOException read or write error.
     */
    public void addString(final String relativeFileName, final String content, final String charset)
            throws IOException
    {
        TarEntry entry = new TarEntry(relativeFileName);
        byte [] bytes = content.getBytes(charset);
        entry.setSize(bytes.length);
        entry.setModTime(System.currentTimeMillis());
        tos.putNextEntry(entry);
        tos.write(bytes);
        tos.closeEntry();
    }

    /**
     * Closes the tar archive.
     */
    public void close()
    {
        if (tos != null)
        {
            IOUtils.closeQuietly(tos);
            tos = null;
        }
        if (bos != null)
        {
            IOUtils.closeQuietly(bos);
            bos = null;
        }
        if (fos != null)
        {
            IOUtils.closeQuietly(fos);
            fos = null;
        }
    }

}
