package it.davideorlandi.jtbtar;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

import com.ice.tar.TarEntry;
import com.ice.tar.TarInputStream;

/**
 * Easy object for extracting from a tar archive.
 * @author Davide
 */
public class EasyUnTar
{
    private static final int BUFFER_SIZE = 1024 * 1024;

    private File baseDirectory;

    private InputStream fis;

    private BufferedInputStream bis;

    private TarInputStream tis;

    /**
     * Constructor.
     * @param baseDirectory base destination directory. Destination file names will be relative to
     *        this location.
     * @param sourceFile source tar archive.
     * @throws IOException error writing the archive.
     */
    public EasyUnTar(final File baseDirectory, final File sourceFile) throws IOException
    {
        this.baseDirectory = baseDirectory;

        boolean ok = false;
        try
        {
            fis = new FileInputStream(sourceFile);
            bis = new BufferedInputStream(fis, BUFFER_SIZE);
            tis = new TarInputStream(bis);
            ok = true;
        }
        finally
        {
            // close streams if partially open
            if ( ! ok)
                close();
        }
    }

    /**
     * Extracts an entry to a file or directory.
     * @return true if an entry was extracted, false if there are no more entries in the archive.
     * @throws IOException read/write error.
     */
    public boolean extractToFile() throws IOException
    {
        TarEntry entry = tis.getNextEntry();
        if (entry == null)
            return false;

        File destFile = new File(baseDirectory, entry.getName());
        if (entry.isDirectory())
        {
            FileUtils.forceMkdir(destFile);
        }
        else
        {
            /*
             * Mobile Atlas Creator doesn't store directories in the tar file, just the files; this
             * not a correct behaviour, but if we want to be able to untar one of its atlases we
             * need to create missing directories.
             */
            File parentDir = destFile.getParentFile();
            if ( ! parentDir.exists())
            {
                FileUtils.forceMkdir(parentDir);
            }

            FileOutputStream fos = null;
            try
            {
                fos = new FileOutputStream(destFile);
                IOUtils.copy(tis, fos);
                fos.close();
            }
            finally
            {
                if (fos == null)
                {
                    IOUtils.closeQuietly(fos);
                    fos = null;
                }
            }
        }

        return true;
    }

    /**
     * Get a list of the names of the remaining entries in the tar archive. After the call, this
     * {@link EasyUnTar} object <b>cannot</b> be reused and <b>it must be closed</b>.
     * @return file names.
     * @throws IOException read/write error.
     */
    public String [] scanFileNames() throws IOException
    {
        ArrayList<String> l = new ArrayList<String>();

        TarEntry entry;
        while ((entry = tis.getNextEntry()) != null)
        {
            l.add(entry.getName());
        }
        return l.toArray(new String [l.size()]);
    }

    /**
     * Closes the tar archive.
     */
    public void close()
    {
        if (tis != null)
        {
            IOUtils.closeQuietly(tis);
            tis = null;
        }
        if (bis != null)
        {
            IOUtils.closeQuietly(bis);
            bis = null;
        }
        if (fis != null)
        {
            IOUtils.closeQuietly(fis);
            fis = null;
        }
    }
}
