package it.davideorlandi.jtbtar;

import it.davideorlandi.jtbtar.exception.AbortedException;
import it.davideorlandi.jtbtar.exception.DisplayebleException;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.filefilter.FileFileFilter;

import com.ice.tar.TarEntry;

/**
 * @author Davide
 */
public class TbtarEngine
{
    private static final String UNPACKED_ATLAS_EXTENSION = ".tba";

    private static final String TAR_EXTENSION = ".tar";

    private static final String [] MAP_EXTENSIONS = new String [] { ".map", ".gmi" };

    private static final String INDEX_EXTENSION = ".tmi";

    private static final String CHARSET = "ISO-8859-1";

    private static final FileFilter MAPFILEIFLTER = new EndsWithFileFilter(MAP_EXTENSIONS);

    private static final FileFilter TARFILEFILTER = new EndsWithFileFilter(TAR_EXTENSION);

    private static final int BUFFER_SIZE = 512 * 1024;

    /**
     * Pack an atlas.
     * @param sourceUnpackedDir source directory.
     * @param destinationTarDir destination diectory.
     * @param includeTilesIndexInRoot if true, ".set" tile index files will be generated and
     *        included in the root archive.
     * @param includeTilesIndexInMap if true, a ".set" tile index file will be generated and
     *        included in each map archive.
     * @param notifier progress notifier.
     * @param clear if true, destination directory will be cleared before packing.
     * @throws IOException read/write error.
     * @throws DisplayebleException generic error (wrong atlas structure etc.).
     */
    public void tar(final File sourceUnpackedDir, final File destinationTarDir,
            final boolean includeTilesIndexInRoot, final boolean includeTilesIndexInMap,
            final ProgressNotifier notifier, final boolean clear) throws IOException,
            DisplayebleException
    {
        checkValidDirectories(sourceUnpackedDir, destinationTarDir);

        AtlasInfo info = getAtlasInfo(sourceUnpackedDir, UNPACKED_ATLAS_EXTENSION);

        // note: number of parts = 1 atlas root + number maps
        int totalOperations = info.totalMaps + 1;

        // clear destination dir if requested
        if (clear)
        {
            clearDirectory(destinationTarDir, notifier);
        }

        File destinationFile = new File(destinationTarDir,
                FilenameUtils.getBaseName(info.atlasName) + TAR_EXTENSION);

        tarRoot(destinationFile, sourceUnpackedDir, info, includeTilesIndexInRoot,
                new PartitionProgressNotifier(notifier, totalOperations, 0));

        int i = 1;
        for (String layer : info.layers.keySet())
        {
            File layerDir = new File(sourceUnpackedDir, layer);
            File destLayerDir = new File(destinationTarDir, layer);

            List<String> maps = info.layers.get(layer);
            for (String map : maps)
            {
                File mapDir = new File(layerDir, map);
                File destMapDir = new File(destLayerDir, map);
                if ( ! destMapDir.exists())
                    FileUtils.forceMkdir(destMapDir);

                ProgressNotifier notifierForMap = new PartitionProgressNotifier(notifier,
                        totalOperations, i);

                notifierForMap.displayProgress(0.0, LocaleManager.format(
                        "ui.progress.subtask.tar.map", layer + "/" + map));
                tarMap(mapDir, destMapDir, includeTilesIndexInMap, notifierForMap, false);

                i++;
            }
        }
    }

    /**
     * Unpack a tared atlas into a directory.
     * @param sourceTarDir tar atlas directory.
     * @param destinationUnpackedDir destination directory.
     * @param notifier progress notifier.
     * @param clear if true, destination directory will be cleared before unpacking.
     * @throws IOException read/write error.
     * @throws DisplayebleException generic error (wrong atlas structure etc.).
     */
    public void unTar(final File sourceTarDir, final File destinationUnpackedDir,
            final ProgressNotifier notifier, final boolean clear) throws IOException,
            DisplayebleException
    {
        checkValidDirectories(sourceTarDir, destinationUnpackedDir);
        AtlasInfo info = getPackedAtlasInfo(sourceTarDir);

        // note: number of parts = 1 atlas root + number maps
        int totalOperations = info.totalMaps + 1;

        // clear destination dir if requested
        if (clear)
        {
            clearDirectory(destinationUnpackedDir, notifier);
        }

        // if tar index file present, unpack it
        if (info.atlasName != null)
        {
            File tarAtlasFile = new File(sourceTarDir, info.atlasName);
            unTarRoot(tarAtlasFile, destinationUnpackedDir, new PartitionProgressNotifier(notifier,
                    totalOperations, 0));
        }

        int i = 1;
        for (String layer : info.layers.keySet())
        {
            File layerDir = new File(sourceTarDir, layer);
            File destLayerDir = new File(destinationUnpackedDir, layer);

            List<String> maps = info.layers.get(layer);
            for (String map : maps)
            {
                File tarMapDir = new File(layerDir, map);
                File destMapDir = new File(destLayerDir, map);
                if ( ! destMapDir.exists())
                    FileUtils.forceMkdir(destMapDir);

                ProgressNotifier notifierForMap = new PartitionProgressNotifier(notifier,
                        totalOperations, i);

                notifierForMap.displayProgress(0.0, LocaleManager.format(
                        "ui.progress.subtask.untar.map", layer + "/" + map));
                unTarMap(tarMapDir, destMapDir, notifierForMap, false);

                i++;
            }
        }
    }

    /**
     * Throw an exception if source or destination directories are empty or not existing or equal.
     * @param srcDir source directory.
     * @param destDir destination directory.
     * @throws DisplayebleException wrong directories.
     */
    private void checkValidDirectories(final File srcDir, final File destDir)
            throws DisplayebleException
    {
        if (srcDir.getName().length() == 0)
            throw new DisplayebleException(
                    LocaleManager.format("error.sourcedirectory.notspecified"));

        if (destDir.getName().length() == 0)
            throw new DisplayebleException(
                    LocaleManager.format("error.destinationdirectory.notspecified"));

        if ( ! srcDir.exists() || ! srcDir.isDirectory())
            throw new DisplayebleException(LocaleManager.format("error.sourcedirectory.missing",
                    srcDir.getAbsolutePath()));

        if ( ! destDir.exists() || ! destDir.isDirectory())
            throw new DisplayebleException(LocaleManager.format(
                    "error.destinationdirectory.missing", destDir.getAbsolutePath()));

        if (destDir.getAbsoluteFile().equals(srcDir.getAbsoluteFile()))
            throw new DisplayebleException(LocaleManager.format("error.directories.coincide"));
    }

    /**
     * Get information about an atlas.
     * @param srcDir atlas directory.
     * @param extension estension of the main atlas file.
     * @return atlas information.
     * @throws DisplayebleException generic error.
     */
    private AtlasInfo getAtlasInfo(final File srcDir, final String extension)
            throws DisplayebleException
    {
        // get atlas name
        File [] files = srcDir.listFiles(new EndsWithFileFilter(UNPACKED_ATLAS_EXTENSION));
        if (files.length < 1)
        {
            throw new DisplayebleException(LocaleManager.format("error.no.unpacked.atlas",
                    UNPACKED_ATLAS_EXTENSION, srcDir.getAbsoluteFile()));
        }

        if (files.length > 1)
        {
            throw new DisplayebleException(LocaleManager.format("error.multiple.unpacked.atlases",
                    UNPACKED_ATLAS_EXTENSION, srcDir.getAbsoluteFile()));
        }
        String baseFullName = files[0].getName();

        // get maps
        File [] layers = srcDir.listFiles();
        Map<String, List<String>> layersMap = new TreeMap<String, List<String>>();
        int counter = 0;
        for (File layer : layers)
        {
            if (layer.isDirectory())
            {
                File [] maps = layer.listFiles();
                List<String> mapNames = new ArrayList<String>();
                for (File map : maps)
                {
                    if (map.isDirectory())
                    {
                        mapNames.add(map.getName());
                        counter++;
                    }
                }
                layersMap.put(layer.getName(), mapNames);
            }
        }

        return new AtlasInfo(baseFullName, layersMap, counter);
    }

    /**
     * Get information about a packed atlas.
     * @param tarDir atlas directory.
     * @return atlas information; atlas name may be null if no tared index is present.
     * @throws DisplayebleException generic error.
     */
    private AtlasInfo getPackedAtlasInfo(final File tarDir) throws DisplayebleException
    {
        // get atlas name
        String atlasName = findTarFile(tarDir, false);

        // get maps
        File [] layers = tarDir.listFiles();
        Map<String, List<String>> layersMap = new TreeMap<String, List<String>>();
        int counter = 0;
        for (File layer : layers)
        {
            if (layer.isDirectory())
            {
                File [] maps = layer.listFiles();
                List<String> mapNames = new ArrayList<String>();
                for (File map : maps)
                {
                    if (map.isDirectory())
                    {
                        mapNames.add(map.getName());
                        counter++;
                    }
                }
                layersMap.put(layer.getName(), mapNames);
            }
        }

        return new AtlasInfo(atlasName, layersMap, counter);
    }

    /**
     * Pack the main atlas file (.tba file + .map files + set directories).
     * @param destinationFile destination archive.
     * @param sourceDirectory source directory.
     * @param atlasInfo atlas information.
     * @param includeTilesIndexes if true, ".set" index filea will be included the archive.
     * @param notifier progress notifier for the tar operation.
     * @throws IOException read/write error.
     * @throws DisplayebleException wrong atlas format.
     * @throws AbortedException user aborts operation.
     */
    private void tarRoot(final File destinationFile, final File sourceDirectory,
            final AtlasInfo atlasInfo, final boolean includeTilesIndexes,
            final ProgressNotifier notifier) throws IOException, DisplayebleException
    {
        EasyTar tar = new EasyTar(sourceDirectory, destinationFile);

        try
        {
            tar.addFile(atlasInfo.atlasName);
            int i = 0;
            for (String layerName : atlasInfo.layers.keySet())
            {
                // layer (dir)
                tar.addFile(layerName);
                List<String> mapNames = atlasInfo.layers.get(layerName);
                for (String mapName : mapNames)
                {
                    if ( ! notifier.displayProgress(100.0 * i / atlasInfo.totalMaps, null))
                    {
                        throw new AbortedException();
                    }

                    String layerAndMap = layerName + "/" + mapName;
                    // layer/map (dir)
                    tar.addFile(layerAndMap);
                    File mapDir = new File(sourceDirectory, layerAndMap);

                    String calibrationFile = findCalibrationFile(mapDir);

                    // layer/map/calibration file
                    tar.addFile(layerAndMap + "/" + calibrationFile);

                    if (includeTilesIndexes)
                    {
                        // layer/map/xxx.set file
                        tar.addString(layerAndMap + "/"
                                + FilenameUtils.getBaseName(calibrationFile) + ".set",
                                tilesToString(listTiles(mapDir)), CHARSET);
                    }

                    // layer/map/set (dir)
                    tar.addFile(layerAndMap + "/set");

                    i++;
                }
            }
        }
        finally
        {
            tar.close();
        }
    }

    /**
     * Unpack everything in the main atlas file.
     * @param sourceFile source archive.
     * @param destinationDirectory destnation directory.
     * @param notifier progress notifier for the untar operation.
     * @throws IOException read/write error.
     * @throws AbortedException user aborts operation.
     */
    private void unTarRoot(final File sourceFile, final File destinationDirectory,
            final ProgressNotifier notifier) throws IOException, AbortedException
    {
        EasyUnTar untar = new EasyUnTar(destinationDirectory, sourceFile);

        try
        {
            boolean stillMoreEntries = false;
            do
            {
                // we don't know the number of files in the archive... so it's always 0%
                // (but we DO check for abort request)
                if ( ! notifier.displayProgress(0.0, null))
                {
                    throw new AbortedException();
                }

                stillMoreEntries = untar.extractToFile();

            } while (stillMoreEntries);
        }
        finally
        {
            untar.close();
        }
    }

    /**
     * Convert a tile list into a string (newline-separated).
     * @param tiles list of tiles.
     * @return newline-separated string.
     */
    private String tilesToString(final File [] tiles)
    {
        // start with a large enough StringBuilder (faster)
        StringBuilder s = new StringBuilder(tiles.length * 16);
        String newLine = System.getProperty("line.separator");
        for (File tile : tiles)
        {
            s.append(tile.getName());
            s.append(newLine);
        }
        return s.toString();
    }

    /**
     * List the tiles inside the "set" directory of the map directory.
     * @param mapDir map directory.
     * @return tiles.
     * @throws DisplayebleException set dir not found.
     */
    private File [] listTiles(final File mapDir) throws DisplayebleException
    {
        File setDir = new File(mapDir, "set");
        if (( ! setDir.exists()) || ( ! setDir.isDirectory()))
        {
            throw new DisplayebleException(LocaleManager.format("error.no.set.dir",
                    mapDir.getAbsoluteFile()));
        }
        return setDir.listFiles((FileFilter) FileFileFilter.FILE);
    }

    /**
     * Return the relative name of the calibration file.
     * @param mapDir map directory.
     * @return calibration file name.
     * @throws DisplayebleException more than one calibration file is present, or none at all.
     */
    private String findCalibrationFile(final File mapDir) throws DisplayebleException
    {
        File [] mapCalibrationFiles = mapDir.listFiles(MAPFILEIFLTER);
        if (mapCalibrationFiles.length < 1)
        {
            throw new DisplayebleException(LocaleManager.format("error.no.calibration.file",
                    mapDir.getAbsoluteFile()));
        }

        if (mapCalibrationFiles.length > 1)
        {
            throw new DisplayebleException(LocaleManager.format("error.multiple.calibration.files",
                    mapDir.getAbsoluteFile()));
        }
        return mapCalibrationFiles[0].getName();
    }

    /**
     * Return the relative name of a tar file inside the specified directory.
     * @param dir directory to be searched.
     * @param ignoreNotFound true: if no tar file found, this method will just return null. false:
     *        if no tar file found, the method will throw an exception.
     * @return tar file name.
     * @throws DisplayebleException more than one tar file, or none at all.
     */
    private String findTarFile(final File dir, final boolean ignoreNotFound)
            throws DisplayebleException
    {
        File [] files = dir.listFiles(TARFILEFILTER);
        if (files.length < 1)
        {
            if (ignoreNotFound)
            {
                return null;
            }
            else
            {
                throw new DisplayebleException(LocaleManager.format("error.no.tar.file",
                        dir.getAbsoluteFile()));
            }
        }

        if (files.length > 1)
        {
            throw new DisplayebleException(LocaleManager.format("error.multiple.tar.files",
                    dir.getAbsoluteFile()));
        }
        return files[0].getName();
    }

    /**
     * Pack a single map (tiles and calibration file), adding a ".tmi" index.
     * @param destinationTarDir destination directory.
     * @param sourceMapDir source map directory.
     * @param includeTilesIndex if true, a .set file will be included in the packed map.
     * @param notifier progress notifer for the tar operation.
     * @param clear if true, destination directory will be cleared before packing.
     * @throws IOException read/write error.
     * @throws DisplayebleException map format error.
     */
    public void tarMap(final File sourceMapDir, final File destinationTarDir,
            final boolean includeTilesIndex, final ProgressNotifier notifier, final boolean clear)
            throws IOException, DisplayebleException
    {
        File [] tiles = listTiles(sourceMapDir);

        // clear destination dir if requested
        if (clear)
        {
            clearDirectory(destinationTarDir, notifier);
        }

        File destinationFile = new File(destinationTarDir,
                FilenameUtils.getBaseName(destinationTarDir.getName()) + ".tar");

        EasyTar tar = new EasyTar(sourceMapDir, destinationFile);
        try
        {
            String calibrationFile = findCalibrationFile(sourceMapDir);
            // calibration file
            tar.addFile(calibrationFile);

            // .set index file
            if (includeTilesIndex)
            {
                tar.addString(FilenameUtils.getBaseName(calibrationFile) + ".set",
                        tilesToString(tiles), CHARSET);
            }
            // set directory
            tar.addFile("set");

            // all the tiles
            int i = 0;
            for (File tile : tiles)
            {
                tar.addFile("set/" + tile.getName());
                if ( ! notifier.displayProgress(100.0 * i / tiles.length, null))
                {
                    throw new AbortedException();
                }
                i++;
            }
        }
        finally
        {
            tar.close();
        }

        // create .tmi index
        File tarIndexFile = new File(
                FilenameUtils.removeExtension(destinationFile.getAbsolutePath()) + INDEX_EXTENSION);
        createTarIndex(destinationFile, tarIndexFile);
    }

    /**
     * Unpack a tared map into a destination directory.
     * @param sourceTarDir packed map dir containing the file to be unpacked.
     * @param destinationMapDir destination directory.
     * @param notifier progress notifier.
     * @param clear if true, destination directory will be cleared before unpacking.
     * @throws IOException read/write error.
     * @throws DisplayebleException generic error (wrong atlas structure etc.).
     * @throws AbortedException user requested abort.
     */
    public void unTarMap(final File sourceTarDir, final File destinationMapDir,
            final ProgressNotifier notifier, final boolean clear) throws IOException,
            DisplayebleException
    {
        // clear destination dir if requested
        if (clear)
        {
            clearDirectory(destinationMapDir, notifier);
        }

        File tarMapArchiveFile = new File(sourceTarDir, findTarFile(sourceTarDir, false));

        String fileNames[];
        EasyUnTar untar = new EasyUnTar(destinationMapDir, tarMapArchiveFile);

        // list archive contents
        try
        {
            fileNames = untar.scanFileNames();
        }
        finally
        {
            untar.close();
        }

        int numFiles = fileNames.length;

        // re-open tha tar archive
        untar = new EasyUnTar(destinationMapDir, tarMapArchiveFile);

        try
        {
            int i = 0;
            boolean stillMoreEntries = false;
            do
            {
                stillMoreEntries = untar.extractToFile();
                if ( ! notifier.displayProgress(100.0 * i / numFiles, null))
                {
                    throw new AbortedException();
                }
                i++;

            } while (stillMoreEntries);

        }
        finally
        {
            untar.close();
        }

        // create .set index file
        String calibrationFileName = findCalibrationFile(destinationMapDir);
        File setFile = new File(destinationMapDir, FilenameUtils.getBaseName(calibrationFileName)
                + ".set");
        FileWriter fw = null;
        BufferedWriter bw = null;
        try
        {
            fw = new FileWriter(setFile);
            bw = new BufferedWriter(fw);
            for (String fileName : fileNames)
            {
                // include files starting with "set/" but not the directory itself
                if (fileName.toLowerCase().startsWith("set/") && (fileName.length() > 4))
                {
                    bw.write(fileName.substring(4));
                    bw.newLine();
                }
            }
            bw.close();
        }
        finally
        {
            if (bw != null)
            {
                IOUtils.closeQuietly(bw);
                bw = null;
            }

            if (fw != null)
            {
                IOUtils.closeQuietly(fw);
                fw = null;
            }
        }

    }

    /**
     * Creates an index file from a tar archive, such as with "-tR" options in tar.
     * @param tarFile tar file to scan.
     * @param indexFile index file to create.
     * @throws IOException read or write errot.
     */
    private void createTarIndex(final File tarFile, final File indexFile) throws IOException
    {
        FileInputStream fis = null;
        FileWriter fw = null;
        BufferedInputStream bis = null;
        BufferedWriter bw = null;
        PrintWriter pw = null;
        IndexedTarInputStream tis = null;
        try
        {
            fis = new FileInputStream(tarFile);
            bis = new BufferedInputStream(fis, BUFFER_SIZE);
            fw = new FileWriter(indexFile);
            bw = new BufferedWriter(fw, BUFFER_SIZE);
            pw = new PrintWriter(bw);
            tis = new IndexedTarInputStream(bis);
            TarEntry entry = null;
            while ((entry = tis.getNextEntry()) != null)
            {
                pw.println(formatIndexRow(entry.getName(), tis.getAbsoluteRecordNumber()));
            }
            pw.println(formatIndexRow(null, tis.getAbsoluteRecordNumber()));

            pw.close();
            pw = null;
        }
        finally
        {
            if (tis != null)
                IOUtils.closeQuietly(tis);
            if (bis != null)
                IOUtils.closeQuietly(bis);
            if (fis != null)
                IOUtils.closeQuietly(fis);
            if (pw != null)
                IOUtils.closeQuietly(pw);
            if (bw != null)
                IOUtils.closeQuietly(bw);
            if (fw != null)
                IOUtils.closeQuietly(fw);
        }
    }

    /**
     * Format a line in an index file. <br>
     * The format is: "block " + right aligned block position (10 characters) + ": " + file name
     * @param name file name or null; if null the string "** Block of NULs **" will be used.
     * @param recordPosition block position.
     * @return formatted string.
     */
    private String formatIndexRow(final String name, final long recordPosition)
    {
        StringBuilder buf = new StringBuilder(64);
        String stringRecordPosition = Long.toString(recordPosition);
        buf.append("block ");
        for (int i = 0; i < 10 - stringRecordPosition.length(); i++)
            buf.append(' ');
        buf.append(stringRecordPosition);
        buf.append(": ");
        buf.append(name != null ? name : "** Block of NULs **");
        return buf.toString();
    }

    /**
     * Delete the contents of the specified directory.
     * @param dir directory to be emptied.
     * @param notifier progress notifier; progress bar will stick to 0%, only the notes will be
     *        updated.
     * @throws IOException read/write error.
     */
    private void clearDirectory(final File dir, final ProgressNotifier notifier) throws IOException
    {
        File [] files = dir.listFiles();

        for (File file : files)
        {
            notifier.displayProgress(0.0, LocaleManager.format("ui.progress.subtask.clear.delete",
                    file.getName()));
            FileUtils.forceDelete(file);
        }
    }

    /**
     * Inforation about an atlas.
     * @author Davide
     */
    private class AtlasInfo
    {
        /** name of the atlas (atlas main file, without path) */
        public String atlasName;

        /**
         * map of layers; the key is the relative name of the layer, the value si a list of relative
         * map names
         */
        public Map<String, List<String>> layers;

        /**
         * number of maps inside all layers
         */
        public int totalMaps;

        /**
         * Constructor.
         * @param atlasName atlas name with extension.
         * @param layers map of layers (see {@link #layers}).
         * @param totalMaps number of maps.
         */
        public AtlasInfo(final String atlasName, final Map<String, List<String>> layers,
                final int totalMaps)
        {
            this.atlasName = atlasName;
            this.layers = layers;
            this.totalMaps = totalMaps;
        }

    }

    private static class EndsWithFileFilter implements FileFilter
    {
        private String [] extensions;

        /**
         * Builds an {@link EndsWithFileFilter} matching a single extension.
         * @param extension accepted extension.
         */
        public EndsWithFileFilter(final String extension)
        {
            this(new String [] { extension });
        }

        /**
         * Builds an {@link EndsWithFileFilter} matching a multiple extensions.
         * @param extensions accepted extensions.
         */
        public EndsWithFileFilter(final String [] extensions)
        {
            this.extensions = extensions;
        }

        @Override
        public boolean accept(final File pathname)
        {
            for (int i = 0; i < extensions.length; i++)
            {
                if (pathname.getName().endsWith(extensions[i]))
                    return true;
            }
            return false;
        }
    }
}
