package it.davideorlandi.jtbtar.ui;

import it.davideorlandi.jtbtar.LocaleManager;
import it.davideorlandi.jtbtar.ProgressNotifier;
import it.davideorlandi.jtbtar.TbtarEngine;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.prefs.Preferences;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

import net.miginfocom.swing.MigLayout;

/**
 * UI panel to pack/unpack atlases.
 * @author Davide
 */
public class AtlasPanel extends JPanel implements CanBeDisabled
{
    /** source path */
    final JTextField sourcePath;

    /** destination path */
    final JTextField destinationPath;

    /** button: pack atlas */
    final JButton tarButton;

    /** button: unpack atlas */
    final JButton untarButton;

    /** button: browse for source atlas dir */
    final JButton sourcePathBrowseButton;

    /** button: browse for destination atlas dir */
    final JButton destinationPathBrowseButton;

    /** checkbox: include .set in root archive */
    final JCheckBox includeSetIndexInRootCheckBox;

    /** checkbox: include .set in map archive */
    final JCheckBox includeSetIndexInMapCheckBox;

    /**
     * Constructor.
     * @param engine application engine.
     */
    public AtlasPanel(final TbtarEngine engine)
    {
        super();

        final JFileChooser fileChooser = new JFileChooser();
        final AtlasPanel panel = this;

        sourcePath = new JTextField();
        sourcePath.setColumns(UIConstants.PREFERRED_COLUMNS);

        destinationPath = new JTextField();
        destinationPath.setColumns(UIConstants.PREFERRED_COLUMNS);

        sourcePathBrowseButton = new JButton(LocaleManager.format("ui.browse.button"));
        sourcePathBrowseButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                File dir = UIUtilities.directoryChooser(fileChooser, (Component) e.getSource(),
                        new File(sourcePath.getText()));
                if (dir != null)
                    sourcePath.setText(dir.getAbsolutePath());
            }
        });

        destinationPathBrowseButton = new JButton(LocaleManager.format("ui.browse.button"));
        destinationPathBrowseButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                File dir = UIUtilities.directoryChooser(fileChooser, (Component) e.getSource(),
                        new File(destinationPath.getText()));
                if (dir != null)
                    destinationPath.setText(dir.getAbsolutePath());
            }
        });

        tarButton = new JButton(LocaleManager.format("ui.tar.atlas.button"));
        tarButton.setMnemonic(KeyStroke.getKeyStroke(
                LocaleManager.format("ui.tar.atlas.button.mnemonic")).getKeyCode());
        tarButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                Component parent = (Component) e.getSource();
                try
                {
                    setControlsEnabled(false);
                    final File srcDir = new File(sourcePath.getText());
                    final File destDir = new File(destinationPath.getText());
                    if ( ! UIUtilities.checkDestinationDirectoryExists(destDir, parent))
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    final Boolean clear = UIUtilities.checkEmptyDestinationDirectory(destDir,
                            parent);
                    if (clear == null)
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    ProgressNotifier notifier = new UIProgressNotifier(LocaleManager.format(
                            "ui.progress.task.tar.atlas", srcDir.getPath(), destDir.getPath(),
                            srcDir.getName(), destDir.getName()), parent, panel);
                    LongRunningTask task = new LongRunningTask(notifier)
                    {
                        @Override
                        public void runnableCode() throws Exception
                        {
                            engine.tar(srcDir, destDir, includeSetIndexInRootCheckBox.isSelected(),
                                    includeSetIndexInMapCheckBox.isSelected(), this.notifier, clear);
                        }
                    };
                    task.start();

                }
                catch (Exception ex)
                {
                    UIUtilities.handleException(ex, (Component) e.getSource());
                }
            }
        });
        untarButton = new JButton(LocaleManager.format("ui.untar.atlas.button"));
        untarButton.setMnemonic(KeyStroke.getKeyStroke(
                LocaleManager.format("ui.untar.atlas.button.mnemonic")).getKeyCode());
        untarButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                Component parent = (Component) e.getSource();
                try
                {
                    setControlsEnabled(false);
                    final File srcDir = new File(sourcePath.getText());
                    final File destDir = new File(destinationPath.getText());
                    if ( ! UIUtilities.checkDestinationDirectoryExists(destDir, parent))
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    final Boolean clear = UIUtilities.checkEmptyDestinationDirectory(destDir,
                            parent);
                    if (clear == null)
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    ProgressNotifier notifier = new UIProgressNotifier(LocaleManager.format(
                            "ui.progress.task.untar.atlas", srcDir.getPath(), destDir.getPath(),
                            srcDir.getName(), destDir.getName()), parent, panel);
                    LongRunningTask task = new LongRunningTask(notifier)
                    {
                        @Override
                        public void runnableCode() throws Exception
                        {
                            engine.unTar(srcDir, destDir, this.notifier, clear);
                        }
                    };
                    task.start();

                }
                catch (Exception ex)
                {
                    UIUtilities.handleException(ex, (Component) e.getSource());
                }
            }
        });

        JLabel unpackedAtlasLabel = new JLabel(LocaleManager.format("ui.source.atlas.label"));
        JLabel packedAtlasLabel = new JLabel(LocaleManager.format("ui.destination.atlas.label"));

        includeSetIndexInRootCheckBox = new JCheckBox(
                LocaleManager.format("ui.atlas.include.set.root.label"));
        includeSetIndexInMapCheckBox = new JCheckBox(
                LocaleManager.format("ui.atlas.include.set.map.label"));

        setLayout(new MigLayout("wrap", "[][grow,fill][]"));
        add(unpackedAtlasLabel);
        add(sourcePath);
        add(sourcePathBrowseButton);
        add(packedAtlasLabel);
        add(destinationPath);
        add(destinationPathBrowseButton);
        add(includeSetIndexInMapCheckBox, "newline, span, split, gapy unrel");
        add(includeSetIndexInRootCheckBox);
        add(tarButton, "newline, tag other, span, split, gapy unrel");
        add(untarButton, "tag other");
    }

    /**
     * Read the preferences for this panel.
     * @param prefs preferences object for the application.
     */
    public void readPreferences(final Preferences prefs)
    {
        sourcePath.setText(prefs.get(PreferenceKeys.SOURCE_ATLAS_DIR, ""));
        destinationPath.setText(prefs.get(PreferenceKeys.DESTINATION_ATLAS_DIR, ""));
        includeSetIndexInRootCheckBox.setSelected(prefs.getBoolean(
                PreferenceKeys.INCLUDE_SET_IN_ROOT, false));
        includeSetIndexInMapCheckBox.setSelected(prefs.getBoolean(
                PreferenceKeys.INCLUDE_SET_IN_MAPS, false));
    }

    /**
     * Writes the preferences for this panel.
     * @param prefs preferences object for the application.
     */
    public void writePreferences(final Preferences prefs)
    {
        prefs.put(PreferenceKeys.SOURCE_ATLAS_DIR, sourcePath.getText());
        prefs.put(PreferenceKeys.DESTINATION_ATLAS_DIR, destinationPath.getText());
        prefs.putBoolean(PreferenceKeys.INCLUDE_SET_IN_ROOT,
                includeSetIndexInRootCheckBox.isSelected());
        prefs.putBoolean(PreferenceKeys.INCLUDE_SET_IN_MAPS,
                includeSetIndexInMapCheckBox.isSelected());
    }

    /**
     * Enable or disable controls.
     * @param enabled true=enabled, false=disabled.
     */
    @Override
    public void setControlsEnabled(final boolean enabled)
    {
        tarButton.setEnabled(enabled);
        untarButton.setEnabled(enabled);
        destinationPath.setEnabled(enabled);
        sourcePath.setEnabled(enabled);
        destinationPathBrowseButton.setEnabled(enabled);
        sourcePathBrowseButton.setEnabled(enabled);
        includeSetIndexInMapCheckBox.setEnabled(enabled);
        includeSetIndexInRootCheckBox.setEnabled(enabled);
    }

}
