package it.davideorlandi.jtbtar.ui;

import it.davideorlandi.jtbtar.LocaleManager;
import it.davideorlandi.jtbtar.ProgressNotifier;
import it.davideorlandi.jtbtar.TbtarEngine;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.prefs.Preferences;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

import net.miginfocom.swing.MigLayout;

/**
 * UI panel to pack/unpack single maps.
 * @author Davide
 */
public class MapPanel extends JPanel implements CanBeDisabled
{

    /** source path */
    final JTextField sourceMapPath;

    /** destination path */
    final JTextField destinationMapPath;

    /** button: pack map */
    final JButton tarButton;

    /** button: unpack map */
    final JButton untarButton;

    /** button: browse for source map dir */
    final JButton sourceMapPathBrowseButton;

    /** button: browse for destination map dir */
    final JButton destinationMapPathBrowseButton;

    /** checkbox: include .set in map archive */
    final JCheckBox includeSetIndexInMapCheckBox;

    /**
     * Constructor.
     * @param engine application engine.
     */
    public MapPanel(final TbtarEngine engine)
    {
        super();

        final JFileChooser fileChooser = new JFileChooser();
        final MapPanel panel = this;

        sourceMapPath = new JTextField();
        sourceMapPath.setColumns(UIConstants.PREFERRED_COLUMNS);

        destinationMapPath = new JTextField();
        destinationMapPath.setColumns(UIConstants.PREFERRED_COLUMNS);

        sourceMapPathBrowseButton = new JButton(LocaleManager.format("ui.browse.button"));
        sourceMapPathBrowseButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                File dir = UIUtilities.directoryChooser(fileChooser, (Component) e.getSource(),
                        new File(sourceMapPath.getText()));
                if (dir != null)
                    sourceMapPath.setText(dir.getAbsolutePath());
            }
        });

        destinationMapPathBrowseButton = new JButton(LocaleManager.format("ui.browse.button"));
        destinationMapPathBrowseButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                File dir = UIUtilities.directoryChooser(fileChooser, (Component) e.getSource(),
                        new File(destinationMapPath.getText()));
                if (dir != null)
                    destinationMapPath.setText(dir.getAbsolutePath());
            }
        });

        tarButton = new JButton(LocaleManager.format("ui.tar.map.button"));
        tarButton.setMnemonic(KeyStroke.getKeyStroke(
                LocaleManager.format("ui.tar.map.button.mnemonic")).getKeyCode());
        tarButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                Component parent = (Component) e.getSource();
                try
                {
                    setControlsEnabled(false);
                    final File srcDir = new File(sourceMapPath.getText());
                    final File destDir = new File(destinationMapPath.getText());
                    if ( ! UIUtilities.checkDestinationDirectoryExists(destDir, parent))
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    final Boolean clear = UIUtilities.checkEmptyDestinationDirectory(destDir,
                            parent);
                    if (clear == null)
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    ProgressNotifier notifier = new UIProgressNotifier(LocaleManager.format(
                            "ui.progress.task.tar.map", srcDir.getPath(), destDir.getPath(),
                            srcDir.getName(), destDir.getName()), parent, panel);
                    LongRunningTask task = new LongRunningTask(notifier)
                    {
                        @Override
                        public void runnableCode() throws Exception
                        {
                            engine.tarMap(srcDir, destDir,
                                    includeSetIndexInMapCheckBox.isSelected(), this.notifier, clear);
                        }
                    };
                    task.start();

                }
                catch (Exception ex)
                {
                    UIUtilities.handleException(ex, (Component) e.getSource());
                }
            }
        });
        untarButton = new JButton(LocaleManager.format("ui.untar.map.button"));
        untarButton.setMnemonic(KeyStroke.getKeyStroke(
                LocaleManager.format("ui.untar.map.button.mnemonic")).getKeyCode());
        untarButton.addActionListener(new ActionListener()
        {
            @Override
            public void actionPerformed(final ActionEvent e)
            {
                Component parent = (Component) e.getSource();
                try
                {
                    setControlsEnabled(false);
                    final File srcDir = new File(sourceMapPath.getText());
                    final File destDir = new File(destinationMapPath.getText());
                    if ( ! UIUtilities.checkDestinationDirectoryExists(destDir, parent))
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    final Boolean clear = UIUtilities.checkEmptyDestinationDirectory(destDir,
                            parent);
                    if (clear == null)
                    {
                        setControlsEnabled(true);
                        return;
                    }
                    ProgressNotifier notifier = new UIProgressNotifier(LocaleManager.format(
                            "ui.progress.task.untar.map", srcDir.getPath(), destDir.getPath(),
                            srcDir.getName(), destDir.getName()), parent, panel);
                    LongRunningTask task = new LongRunningTask(notifier)
                    {
                        @Override
                        public void runnableCode() throws Exception
                        {
                            engine.unTarMap(srcDir, destDir, this.notifier, clear);
                        }
                    };
                    task.start();

                }
                catch (Exception ex)
                {
                    UIUtilities.handleException(ex, (Component) e.getSource());
                }
            }
        });

        JLabel unpackedMapLabel = new JLabel(LocaleManager.format("ui.source.map.label"));
        JLabel packedMapLabel = new JLabel(LocaleManager.format("ui.destination.map.label"));

        includeSetIndexInMapCheckBox = new JCheckBox(
                LocaleManager.format("ui.map.include.set.map.label"));

        setLayout(new MigLayout("wrap", "[][grow,fill][]"));
        add(unpackedMapLabel);
        add(sourceMapPath);
        add(sourceMapPathBrowseButton);
        add(packedMapLabel);
        add(destinationMapPath);
        add(destinationMapPathBrowseButton);
        add(includeSetIndexInMapCheckBox, "newline, span, split, gapy unrel");
        add(tarButton, "newline, tag other, span, split, gapy unrel");
        add(untarButton, "tag other");

    }

    /**
     * Read the preferences for this panel.
     * @param prefs preferences object for the application.
     */
    public void readPreferences(final Preferences prefs)
    {
        sourceMapPath.setText(prefs.get(PreferenceKeys.SOURCE_MAP_DIR, ""));
        destinationMapPath.setText(prefs.get(PreferenceKeys.DESTINATION_MAP_DIR, ""));
        includeSetIndexInMapCheckBox.setSelected(prefs.getBoolean(
                PreferenceKeys.INCLUDE_SET_IN_MAP, false));
    }

    /**
     * Writes the preferences for this panel.
     * @param prefs preferences object for the application.
     */
    public void writePreferences(final Preferences prefs)
    {
        prefs.put(PreferenceKeys.SOURCE_MAP_DIR, sourceMapPath.getText());
        prefs.put(PreferenceKeys.DESTINATION_MAP_DIR, destinationMapPath.getText());
        prefs.putBoolean(PreferenceKeys.INCLUDE_SET_IN_MAP,
                includeSetIndexInMapCheckBox.isSelected());
    }

    /**
     * Enable or disable controls.
     * @param enabled true=enabled, false=disabled.
     */
    @Override
    public void setControlsEnabled(final boolean enabled)
    {
        tarButton.setEnabled(enabled);
        untarButton.setEnabled(enabled);
        destinationMapPath.setEnabled(enabled);
        sourceMapPath.setEnabled(enabled);
        destinationMapPathBrowseButton.setEnabled(enabled);
        sourceMapPathBrowseButton.setEnabled(enabled);
        includeSetIndexInMapCheckBox.setEnabled(enabled);
    }

}
