package it.davideorlandi.jtbtar.ui;

import it.davideorlandi.jtbtar.LocaleManager;
import it.davideorlandi.jtbtar.exception.AbortedException;
import it.davideorlandi.jtbtar.exception.DisplayebleException;

import java.awt.Component;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.apache.commons.io.FileUtils;

/**
 * Utility methods.
 * @author Davide
 */
public class UIUtilities
{
    /**
     * Display the exception.
     * @param e exception.
     * @param component parent component.
     */
    public static final void handleException(final Throwable e, final Component component)
    {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);

        final String message;
        if (e instanceof DisplayebleException)
            message = e.getMessage();
        else
            message = LocaleManager.format("ui.error.message", e, sw.toString());

        final int type;
        if (e instanceof AbortedException)
            type = JOptionPane.WARNING_MESSAGE;
        else
            type = JOptionPane.ERROR_MESSAGE;

        JOptionPane.showMessageDialog(component, message, LocaleManager.format("ui.error.title"),
                type);

    }

    /**
     * Display a directory chooser.
     * @param fileChooser instantiated file chooser.
     * @param component parent component.
     * @param currentDir current directory to show.
     * @return chosen directory or null if canceled.
     */
    public static File directoryChooser(final JFileChooser fileChooser, final Component component,
            final File currentDir)
    {
        fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        fileChooser.setCurrentDirectory(currentDir);
        int ret = fileChooser.showOpenDialog(component);
        if (ret == JFileChooser.APPROVE_OPTION)
            return fileChooser.getSelectedFile();
        else
            return null;
    }

    /**
     * If the directory is not empty ask the user if he wants to clear it.
     * @param directory directory to be checked.
     * @param component parent component of the dialog.
     * @return true if the user answered yes, false if he answered no (or the directory was empty),
     *         null if he answered cancel.
     * @throws IOException error deleting files.
     */
    public static Boolean checkEmptyDestinationDirectory(final File directory,
            final Component component) throws IOException
    {
        File [] content = directory.listFiles();
        if ((content != null) && (content.length > 0))
        {
            int ret = JOptionPane.showConfirmDialog(component, LocaleManager.format(
                    "ui.directorynotempty.question", directory.getAbsolutePath()),
                    LocaleManager.format("ui.directorynotempty.title"),
                    JOptionPane.YES_NO_CANCEL_OPTION);
            switch (ret)
            {
                case JOptionPane.CANCEL_OPTION:
                    return null;

                case JOptionPane.YES_OPTION:
                    return true;

                case JOptionPane.NO_OPTION:
                    return false;

                default:
                    return null;
            }
        }
        return false;
    }

    /**
     * If the directory doesn't exist ask the user if he wants to create it. And then, if the user
     * answers yes, the directory is created.
     * @param directory directory to be checked.
     * @param component parent component of the dialog.
     * @return true if the user answered ok (or the directory was already there), false if he
     *         answered cancel.
     * @throws IOException error deleting files.
     */
    public static boolean checkDestinationDirectoryExists(final File directory,
            final Component component) throws IOException
    {
        if ( ! directory.exists())
        {
            int ret = JOptionPane.showConfirmDialog(component, LocaleManager.format(
                    "ui.createdirectory.question", directory.getAbsolutePath()),
                    LocaleManager.format("ui.createdirectory.title"), JOptionPane.OK_CANCEL_OPTION);
            switch (ret)
            {
                case JOptionPane.CANCEL_OPTION:
                    return false;

                case JOptionPane.OK_OPTION:
                    FileUtils.forceMkdir(directory);
                    return true;

                default:
                    return false;
            }
        }
        else
        {
            return true;
        }
    }

    /**
     * Find the parent JFrame of a component.
     * @param component component to search.
     * @return parent JFrame.
     * @throws RuntimeException no JFrame found.
     */
    public static final JFrame getParentFrame(final Component component)
    {
        Component current = component;
        while (current != null)
        {
            if (current instanceof JFrame)
                return (JFrame) current;
            current = current.getParent();
        }
        throw new RuntimeException("No JFrame found");
    }
}
